/*
 * usbnix -- *NIX like USB Key ??
 *
 * Author: Koji Suzuki (C) 2007
 *
*/

#ifdef _WIN32
#define WIN32NATIVE
//#define HAVE_TERMIOS
#endif

#ifndef __linux__
//#define TEST1		// echo only
//#define TEST2		// do_login (without do_format)
#define I2CROM_DEBUG
#endif


#ifdef __linux__
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <errno.h>
#include <termios.h>
#include <fcntl.h>

typedef unsigned char uint8_t;
typedef unsigned short uint16_t;
typedef unsigned int uint32_t;

#define PSTR(str)	(str)
#define __LPM(p)	(*p)
#define strcmp_P	strcmp
#define strcpy_P	strcpy

#define O_BINARY	0

#elif defined(WIN32NATIVE)

#include <windows.h>
#include <io.h>
#ifdef HAVE_TERMIOS
#include <termios.h>
#else
#include <wincon.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <errno.h>
#include <fcntl.h>

typedef unsigned char uint8_t;
typedef unsigned short uint16_t;
typedef unsigned int uint32_t;

#define PSTR(str)	(str)
#define __LPM(p)	(*p)
#define strcmp_P	strcmp
#define strcpy_P	strcpy

#else /* AVR */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>
#include <util/delay.h>
#include <stdlib.h>
#include <string.h>

#include "usbcdc.h"
#include "i2c.h"

#endif

#include "minicrypt.h"

#ifndef EXT_BUFSIZE
#define EXT_BUFSIZE	16
#endif

#ifndef EINVAL
#define EINVAL		22
#endif
#ifndef ENOENT
#define ENOENT		 2
#endif
#ifndef ENOSPC
#define ENOSPC		28
#endif
#ifndef EIO
#define EIO		 5
#endif

uint8_t ext_buf[EXT_BUFSIZE];

int8_t i2crom_seek(uint16_t pos);
int8_t i2crom_read16();
int8_t i2crom_write16();

#if defined(__linux__) || defined(WIN32NATIVE)

static int key_fd,rom_fd;
#if defined(WIN32NATIVE) && !defined(HAVE_TERMIOS)
static HANDLE conin;
static DWORD conin_mode_sv;
#else
struct termios term_sv;
#if defined(WIN32NATIVE)
#define tcsetattr(fd, cmd, arg) ioctl(fd, cmd, arg)
#define tcgetattr(fd, buf) ioctl(fd, TCGETA, buf)
#endif
#endif

void lin_term() {
#if defined(WIN32NATIVE) && !defined(HAVE_TERMIOS)
	if (conin_mode_sv != 0) 
		SetConsoleMode(conin, conin_mode_sv);
#else
	tcsetattr(0, TCSANOW, &term_sv);
#endif
	if (key_fd > 0) close(key_fd);
	if (rom_fd > 0) close(rom_fd);
}

void lin_init() {
#if defined(WIN32NATIVE) && !defined(HAVE_TERMIOS)
	DWORD conin_mode = 0;
	int r;

	conin = GetStdHandle(STD_INPUT_HANDLE);
	r = GetConsoleMode(conin, &conin_mode);
	if (r) {
		conin_mode_sv = conin_mode;
		conin_mode &= ~(ENABLE_LINE_INPUT | ENABLE_ECHO_INPUT);
		r = SetConsoleMode(conin, conin_mode);
	} else {
		fprintf(stderr, "run on the console(command prompt)\n");
		exit(1);
	}
#else
	struct termios term;
	tcgetattr(0, &term);
	term_sv = term;
	term.c_lflag &= ~(ECHO | ICANON);
	term.c_iflag &= ~(ICRNL);
	tcsetattr(0, TCSANOW, &term);
#endif
	atexit(lin_term);

	key_fd = open("key.img",O_RDWR|O_BINARY);
	if (key_fd < 0) {
		char buf[128];
#ifdef USE_URANDOM
		key_fd = open("/dev/urandom", O_RDONLY|O_BINARY);
		if (key_fd < 0) {
			printf("can't create key file\n");
			exit(1);
		}
		read(key_fd, buf, 128);
		close(key_fd);
		key_fd = open("key.img",O_RDWR|O_CREAT|O_BINARY,0666);
		if (key_fd < 0) {
			printf("can't create key file\n");
			exit(1);
		}
#else
		key_fd = open("key.img",O_RDWR|O_CREAT|O_BINARY,0666);
		if (key_fd < 0) {
			printf("can't create key file\n");
			exit(1);
		}
		memset(buf, 0xff, 128);
#endif
		write(key_fd, buf, 128);
		memset(buf, 0xff, 16);
		write(key_fd, buf, 16);
		lseek(key_fd, 0L, SEEK_SET);
	}
	rom_fd = open("i2crom.img",O_RDWR|O_BINARY);
	if (rom_fd < 0) {
		char buf[32*1024];
		rom_fd = open("i2crom.img",O_RDWR|O_CREAT|O_BINARY,0666);
		if (rom_fd < 0) {
			printf("can't create rom file\n");
			exit(1);
		}
		memset(buf, 0xff, 32*1024);

		write(rom_fd, buf, 32*1024);
		lseek(rom_fd, 0L, SEEK_SET);
	}
}

int8_t i2crom_seek(uint16_t pos) {
	return lseek(rom_fd, (long)pos, SEEK_SET);
}
int8_t i2crom_read16() {
	return read(rom_fd, ext_buf, 16);
}
int8_t i2crom_write16() {
	return write(rom_fd, ext_buf, 16);
}

void eeprom_read_block(void *, const void *, size_t);
void eeprom_write_block(void *, const void *, size_t);


void eeprom_read_block(void *buf, const void *pos, size_t len) {
	long _pos = (long)(pos);
	lseek(key_fd, _pos, SEEK_SET);
	read(key_fd, buf, len);
}

void eeprom_write_block(void *buf, const void *pos, size_t len) {
	long _pos = (long)(pos);
	lseek(key_fd, _pos, SEEK_SET);
	write(key_fd, buf, len);
}

void usbcdc_putc(char data) {
	write(1, &data, 1);
}

char usbcdc_getc(void) {
	int r;
	char ret = 0;
#if defined(WIN32NATIVE) && !defined(HAVE_TERMIOS)
	char buf[3];
	DWORD nrec = 0;

	while (!ReadConsole(conin, buf, 1, &nrec, NULL) || nrec == 0)
			;
		return buf[0];
#endif
	while ((r = read(0, &ret, 1)) != 1)
		;

	return ret;
}

uint16_t get_seed() {
	uint16_t seed;
	__asm__ __volatile__("rdtsc" : "=A" (seed));
	return seed;
}

#else /* AVR */
static union {
	uint16_t word;
	struct {
		uint8_t lo;
		uint8_t hi;
	};
} i2crom_pos;

static int8_t i2crom_set_pos() {
	int8_t cnt = 0;
	int8_t r;
	uint8_t save[2];
	save[0] = ext_buf[0];
	save[1] = ext_buf[1];
retry:
	cnt++;
	r = i2c_start();
	if (r < 0) {
#if 0
		if ((r == -3) && (cnt < 10)) {
			i2c_stop();
			wdt_reset();
			goto retry;
		}
#endif
		return r;
	}
	ext_buf[0] = i2crom_pos.hi;
	ext_buf[1] = i2crom_pos.lo;
	r = i2c_send(0x50, 2);
	ext_buf[0] = save[0];
	ext_buf[1] = save[1];
	if (r == -2) {
		i2c_stop();
		wdt_reset();
		goto retry;
	}
	return r;
}

int8_t i2crom_seek(uint16_t pos) {
	int8_t r;
	static int called;
	if (!called) {
		init_i2c();
		called = 1;
	}
	i2crom_pos.word = pos;
	r = i2crom_set_pos();
	i2c_stop();
	return r;
}

int8_t i2crom_read16() {
	int8_t r;
retry:
	r = i2c_start();
	if (r < 0) return r;
	r = i2c_recv(0x50, 16);
	i2c_stop();
	if (r == -2) {
		wdt_reset();
		goto retry;
	}
	i2crom_pos.word += 16;
	return r;
}

int8_t i2crom_write16() {
	int8_t r;
retry:
	r = i2crom_set_pos();
	if (r < 0) return r;
	r = i2c_send_cont(16);
	i2c_stop();
	if (r == -2) {
		wdt_reset();
		goto retry;
	}
	i2crom_pos.word += 16;
	return r;
}

static inline void timer_init() {
#if defined(__AVR_ATtiny861__) || defined(__AVR_ATtiny461__)
	TCCR0A = (1<<TCW0); // 16 bit mode
	TCCR0B = 1; // timer0 start @ clk/1
#else
	TCCR1B = 1; // timer1 start @ clk/1
#endif
}

uint16_t get_seed() {
#if defined(__AVR_ATtiny861__) || defined(__AVR_ATtiny461__)
	union {
		uint16_t word;
		struct {
			uint8_t lo;
			uint8_t hi;
		};
	} seed;
	seed.lo = TCNT0L;
	seed.hi = TCNT0H;
	return seed.word;
#else
	return TCNT1;
#endif
}
#endif

#define LINE_BUFSIZE	64
static uint8_t line_buf[LINE_BUFSIZE];

#define MAX_ID			32
#define FILE_CHUNK		1024
#define FILE_CHUNK_SHIFT	10
#define FNAME_SIZE		64

int8_t do_format() {
	uint16_t i;
	uint8_t x;
	int8_t r;

	r = i2crom_seek(0);
	if (r < 0) return -EIO;
	memset(ext_buf, 0xff, 16);
	x = 0;
	for (i = 0; i < FILE_CHUNK/16*MAX_ID; i++) {
		r = i2crom_write16();
		if (r < 0) return -EIO;
		x++;
		if (x >= 64) {
			usbcdc_putc('.');
			x = 0;
		}
	}
	return 0;
}

static void puts_P(const char *p) {
	uint8_t c;
	while ((c = __LPM(p)) != 0) {
		usbcdc_putc(c);
		p++;
	}
}

#define WITH_ECHO	0
#define NO_ECHO		1

static void getln(uint8_t *in, uint8_t len, uint8_t mode) {
	uint8_t ch;

	len--;
	while (len) {
		ch = usbcdc_getc();
		if (mode == WITH_ECHO) {
			if (ch == '\b') {
				if (len != 0) {
					in--;
					len++;
					usbcdc_putc(ch);
					usbcdc_putc(' ');
					usbcdc_putc(ch);
				}
				continue;
			} else {
				usbcdc_putc(ch);
			}
		}
		if (ch == '\n') continue;
		if (ch == '\r') {
			if (mode == NO_ECHO) {
				usbcdc_putc('\r');
			}
			usbcdc_putc('\n');
			break;
		} else {
			*in++ = ch;
		}
		len--;
	}
	*in = 0;
}

static int is_ff16() {
	uint8_t i;
	int8_t ret = 1;
	for (i=0; i<16; i++) {
		if (ext_buf[i] != 0xff) ret = 0;
	}
	return ret;
}

int8_t alloc_id(uint8_t *p) {
	uint8_t i;
	uint16_t pos;
	int8_t len,len2,r;
	uint8_t j;
	len = strlen((char *)p)+1;
	if (len <= 1 || len > LINE_BUFSIZE) return -EINVAL;
	for (i=0,pos=0; i<MAX_ID; i++,pos += FILE_CHUNK) {
		r = i2crom_seek(pos);
		if (r < 0) return -EIO;
		r = i2crom_read16();
		if (r < 0) return -EIO;
		if (is_ff16()) {
			r = i2crom_seek(pos);
			if (r < 0) return -EIO;
			for (j=0;j<4; j++) {
				len2 = len;
				if (len2 > 16) len2 = 16;
				memcpy(ext_buf,p,len2);
				r = i2crom_write16();
				if (r < 0) return -EIO;
				p += len2;
				len -= len2;
				if (len <= 0) {
					break;
				}
			}
			pos += FNAME_SIZE;
			r = i2crom_seek(pos);
			if (r < 0) return -EIO;
			pos = ((i<<4) & (128-1));
			eeprom_read_block(ext_buf, (const void *)pos, 16);
			mc_setkey(ext_buf);
			return i;
		}
	}
	return -ENOSPC;
}

int8_t name_id(uint8_t *p) {
	uint8_t i;
	int8_t len0,len,len2,r;
	uint16_t pos;
	uint8_t *pp;
	len0 = strlen((char *)p)+1;
	if (len0 <= 1 || len0 > LINE_BUFSIZE) return -EINVAL;

	for (i=0,pos=0; i<MAX_ID; i++,pos += FILE_CHUNK) {
		r = i2crom_seek(pos);
		if (r < 0) return -EIO;
		pp = p;
		len = len0;
		for (;;) {
			len2 = len;
			if (len2 > 16) len2 = 16;
			len -= len2;
			r = i2crom_read16();
			if (r < 0) return -EIO;
			if (memcmp(pp,ext_buf,len2)) {
				break;
			}
			if (len <= 0) {
				pos += FNAME_SIZE;
				r = i2crom_seek(pos);
				if (r < 0) return -EIO;
				pos = ((i<<4) & (128-1));
				eeprom_read_block(ext_buf, (const void *)pos, 16);
				mc_setkey(ext_buf);
				return i;
			}
			pp += len2;
		}
	}
	return -ENOENT;
}

int8_t id_exists(int8_t id) {
	uint16_t pos;
	uint8_t i;
	int8_t r;

	if (id<0 || id>= MAX_ID) return -EINVAL;

	pos = id;
	pos <<= FILE_CHUNK_SHIFT;
	r = i2crom_seek(pos);
	if (r < 0) return -EIO;
	r = i2crom_read16();
	if (r < 0) return -EIO;
	if (is_ff16()) return -ENOENT;

	pos += FNAME_SIZE;
	r = i2crom_seek(pos);
	if (r < 0) return -EIO;
	i = id;
	pos = ((i<<4) & (128-1));
	eeprom_read_block(ext_buf, (const void *)pos, 16);
	mc_setkey(ext_buf);
	return 0;
}

int8_t id_remove(int8_t id) {
	uint16_t pos;
	uint8_t i;
	int8_t r;
	if (id < 0 || id >= MAX_ID) return -EINVAL;

	pos = id;
	pos <<= FILE_CHUNK_SHIFT;
	r = i2crom_seek(pos);
	if (r < 0) return -EIO;
	r = i2crom_read16();
	if (r < 0) return -EIO;
	if (is_ff16()) return -ENOENT;
	memset(ext_buf,0xff,16);
	r = i2crom_seek(pos);
	if (r < 0) return -EIO;
	r = i2crom_write16();
	if (r < 0) return -EIO;

	pos += FNAME_SIZE;
	r = i2crom_seek(pos);
	if (r < 0) return -EIO;
	for (i=FNAME_SIZE/16; i<FILE_CHUNK/16; i++) {
		r = i2crom_write16();
		if (r < 0) return -EIO;
	}
	return 0;
}

static void print_int(uint8_t n) {
	uint8_t i,x;
	char buf[4];
	buf[0] = ' ';
	buf[1] = ' ';
	buf[2] = '0';
	buf[3] = ' ';
	for (i=2,x=200; x>0; i--,x-=100) {
		if (n >= x) {
			buf[0] = '0' + i;
			n -= x;
			break;
		}
	}
	for (i=9,x=90; x>0; i--,x-=10) {
		if (n >= x) {
			buf[1] = '0' + i;
			n -= x;
			break;
		}
	}
	for (i=9; i>0; i--) {
		if (n >= i) {
			buf[2] = '0' + i;
			break;
		}
	}
	usbcdc_putc(buf[0]);
	usbcdc_putc(buf[1]);
	usbcdc_putc(buf[2]);
	usbcdc_putc(buf[3]);
}

void print_error(int8_t err) {
	if (err == -EINVAL) {
		puts_P(PSTR("Invalid argument\r\n"));
	} else if (err == -ENOENT) {
		puts_P(PSTR("No such file or directory\r\n"));
	} else if (err == -ENOSPC) {
		puts_P(PSTR("No space left on device\r\n"));
	} else if (err == -EIO) {
		puts_P(PSTR("I/O error\r\n"));
	}
}

int8_t print_filelist(uint8_t *p) {
    uint8_t imode = 0;
    int8_t i,r;
    uint8_t j,ch;
    uint16_t pos;
    if (!strcmp_P((char *)p, PSTR("-i"))) {
	    imode = 1;
	    *p = 0;
    }
    if (p[0] == 0 ||  (p[0] == '*' && p[1] == 0)) { // show all 
	for (i=0,pos=0; i<MAX_ID; i++,pos += FILE_CHUNK) {
		r = i2crom_seek(pos);
		if (r < 0) return -EIO;
		r = i2crom_read16();
		if (r < 0) return -EIO;
		if (is_ff16()) continue;
		if (imode) print_int(i);
		for (;;) {
			for (j=0; j<16; j++) {
				ch = ext_buf[j];
				if (!ch) break;
				usbcdc_putc(ch);
			}
			if (j!=16) {
				usbcdc_putc('\r');
				usbcdc_putc('\n');
				break;
			}
			r = i2crom_read16();
			if (r < 0) return -EIO;
		}
	}
    } else {
	i = name_id(p);
	if (i < 0) {
		return i;
	}
	while (*p) {
		usbcdc_putc(*p);
		p++;
	}
	usbcdc_putc('\r');
	usbcdc_putc('\n');
    }
    return 0;
}

int8_t remove_file(uint8_t *p) {
	int8_t r,i;
	i = name_id(p);
	if (i >= 0) {
		r = id_remove(i);
		if (r < 0) return r;
	} else if (p[0] == '-') {
		uint8_t *pp = p+1;
		while (*pp) {
			if (*pp < '0' || '9' < *pp) {
				return -EINVAL;
			}
			pp++;
		}
		i = atoi((char *)p+1);
		r = id_exists(i);
		if (r) return r;
		r = id_remove(i);
		if (r) return r;
	}
	return i;
}

int8_t create_file(uint8_t *p) {
	int8_t i,end_flag,r;
	uint8_t ch,pch,len2;
	uint16_t len;
	end_flag = 0;
	i = name_id(p);
	if (i < 0) {
		i = alloc_id(p);
		if (i < 0) {
			return i;
		}
	}
	len = 0;
	len2 = 0;
	pch = 0xff;
	for (;;) {
		ch = usbcdc_getc();
		if ((len >= FILE_CHUNK - FNAME_SIZE -1) || (ch == 0)
			|| (ch == ('D' & 0x1f))
			|| (pch == '\n' && ch == '.')) {
			ch = 0;
			end_flag = 1;
		} else {
			if (ch == '\r') {
				usbcdc_putc(ch);
				ch = '\n';
				usbcdc_putc(ch);
			} else {
				usbcdc_putc(ch); // echo back
			}
		}
		ext_buf[len2] = ch;
		len2 ++;
		len++;
		if (end_flag || (len2 >= 16)) {
			mc_encrypt(ext_buf);
			r = i2crom_write16();
			if (r < 0) return -EIO;
			len2 = 0;
		}
		if (end_flag) break;
		pch = ch;
	}
	return i;
}

int8_t print_file(uint8_t *p) {
	int8_t r,i,j,end_flag;
	uint8_t ch;
	uint16_t len;

	end_flag = 0;
	if (p[0] == '-') {
		uint8_t *pp = p+1;
		while (*pp) {
			if (*pp < '0' || '9' < *pp) return -EINVAL;
			pp++;
		}
		i = atoi((char *)p+1);
		r = id_exists(i);
		if (r) return r;
	} else {
		i = name_id(p);
	}
	if (i < 0) {
		return i;
	}
	len = 0;
	while (!end_flag) {
		r = i2crom_read16();
		if (r < 0) return -EIO;
		mc_decrypt(ext_buf);
		for (j=0; j<16; j++) {
			ch = ext_buf[j];
			if (ch == '\n') {
				usbcdc_putc('\r');
			}
			usbcdc_putc(ch);
			len++;
			if ((len >= FILE_CHUNK - FNAME_SIZE) || (ch == 0)) {
				end_flag = 1;
				break;
			}
		}
	}
	return i;
}

#define MAX_USER	16
#define MAX_PASSWORD	16

int8_t do_login() {
	uint8_t len,len2,i,c;
	uint8_t first;
	uint16_t pos;
	
#ifdef TEST2
	puts_P(PSTR("test2\r\n"));
#endif
	eeprom_read_block(ext_buf, (const void *)128, 16);
	first = is_ff16();
	eeprom_read_block(ext_buf, (const void *)0, 16);
	if (is_ff16()) {
		puts_P(PSTR("Type any key 64 times to create key:\r\n"));
		len = 0;
		c = '0';
		for (i=0; i< 128; i+=2) {
			usbcdc_putc(c);
			if (c == '9') c = '0';
			else c++;
		}
		usbcdc_putc('\r');
		usbcdc_putc('\n');
		for (i=0; i< 128; i+=2) {
			usbcdc_getc();
			usbcdc_putc('.');

			pos = get_seed();
			eeprom_write_block(&pos, (void *)(uint16_t)i, 2);
		}
		usbcdc_putc('\r');
		usbcdc_putc('\n');
#if !defined(TEST2)
		puts_P(PSTR("formatting rom "));
		do_format();
		puts_P(PSTR("done\r\n"));
#endif
	}
	puts_P(PSTR("login: "));
	getln(line_buf, MAX_USER, WITH_ECHO);
	len = strlen((char *)line_buf);
	if (len == 0) return 0;

	if (first) {
		puts_P(PSTR("New password: "));
		getln(line_buf+MAX_USER, MAX_PASSWORD, NO_ECHO);
		puts_P(PSTR("Retype new password: "));
		getln(ext_buf, MAX_PASSWORD, NO_ECHO);
		if (strcmp((char *)ext_buf, (char *)line_buf+MAX_USER)) return 0;
	} else {
		puts_P(PSTR("Password: "));
		getln(line_buf+MAX_USER, MAX_PASSWORD, NO_ECHO);
	}
	len2 = strlen((char *)line_buf+MAX_USER);
	for (i=len; i<MAX_USER; i++) {
		line_buf[i] = 0;
	}
	for (i=len2; i<MAX_PASSWORD; i++) {
		line_buf[MAX_USER+i] = 0;
	}
#ifdef TEST2
	puts_P(PSTR("end input\r\n"));
	for (i=0; i<len; i++) {
		usbcdc_putc(line_buf[i]);
	}
	puts_P(PSTR("\r\n"));
	usbcdc_getc();
#endif
	mc_setkey(line_buf);
	mc_encrypt(line_buf+MAX_USER);
	for (pos=0; pos<128; pos+=16) {
		eeprom_read_block(line_buf, (const void *)pos, 16);
		mc_encrypt(line_buf);
	}
#ifdef TEST2
	puts_P(PSTR("end encrypt\r\n"));
	usbcdc_getc();
#endif
	if (first) {
		eeprom_write_block(line_buf, (void *)128, 16);
		i = 1;
	} else {
		eeprom_read_block(ext_buf, (const void *)128, 16);
		i = !memcmp(ext_buf, line_buf, 16);
	}
	if (i) {
		strcpy_P((char *)line_buf, PSTR("motd"));
		if (name_id(line_buf) >= 0) {
			print_file(line_buf);
		}
	}
	return i;
}

#ifdef I2CROM_DEBUG
static int8_t nibble2hex(int8_t n) {
	n &= 0xf;
	if (n < 10) return '0' + n;
	return 'a' + n - 10;
}

int8_t i2crom_dump(uint8_t *p) {
	int8_t r,i;
	union {
		uint16_t word;
		struct {
			uint8_t lo;
			uint8_t hi;
		};
	} pos;

	if (!p || (*p == 0)) pos.word = 0;
	else pos.word = atoi((char *)p);

	i2crom_seek(pos.word);
	memset(ext_buf,0x00,16);
	r = i2crom_read16();
	if (r != 16) {
		puts_P(PSTR("read err -"));
		if (r < 0) {
			usbcdc_putc('0' - r);
		} else {
			usbcdc_putc(nibble2hex(r>>4));
			usbcdc_putc(nibble2hex(r));
		}
		puts_P(PSTR("\r\n"));
	}
	usbcdc_putc(nibble2hex(pos.hi>>4));
	usbcdc_putc(nibble2hex(pos.hi));
	usbcdc_putc(nibble2hex(pos.lo>>4));
	usbcdc_putc(nibble2hex(pos.lo));
	usbcdc_putc(':');
	for (i=0; i<16; i++) {
		usbcdc_putc(nibble2hex(ext_buf[i]>>4));
		usbcdc_putc(nibble2hex(ext_buf[i]));
		usbcdc_putc(' ');
	}
	puts_P(PSTR("\r\n"));
	return 0;
}

#endif
int8_t i2crom_format(uint8_t *p) {
	int8_t r;
retry:
	puts_P(PSTR("Format ? (yes or no) "));
	getln(line_buf, EXT_BUFSIZE, WITH_ECHO);
	if (strcmp((char *)line_buf, "yes")) {
		if (!strcmp((char *)line_buf, "no")) return 0;
		puts_P(PSTR("Please answer yes or no. "));
		goto retry;
	}
	puts_P(PSTR("formatting "));
	r = do_format();
	if (r < 0) return -EIO;
	puts_P(PSTR("done\r\n"));
	return 0;
}

int main() {
	int8_t r,len;
	int8_t create_mode = 0;
	int8_t login_ok = 0;
	uint8_t *p,*cmd_end,*fname;
	char *cmd;

#if defined(__linux__) || defined(WIN32NATIVE)
	lin_init();
#else /* AVR */
#ifndef NO_CLKPR_SET
	/* set system_clock pre scaler 1/8 => 1/1 */
	CLKPR = (1<<CLKPCE);
	CLKPR = 0;
#endif

	timer_init();
	wdt_enable(WDTO_1S);
	usbcdc_init();

	sei();

	usbcdc_getc();
#endif
	for (;;) {
#if defined(__linux__) || defined(WIN32NATIVE)
#else /* AVR */
		wdt_reset();
		usbcdc_poll();
#endif

#if defined(TEST1)
		r = usbcdc_getc();
		if (r == '\n') {
		} else if (r == '\r') {
			usbcdc_putc(r);
			usbcdc_putc('\n');
		} else {
			usbcdc_putc(r);
		}
#else
		create_mode = 0;
		if (!login_ok) {
			login_ok = do_login();
			continue;
		}
		usbcdc_putc('#');
		usbcdc_putc(' ');
		getln(line_buf, LINE_BUFSIZE, WITH_ECHO);

		line_buf[LINE_BUFSIZE-1] = 0;
		len = strlen((char *)line_buf);
		p = line_buf;
		while (*p) {
			if (!(*p == ' ' || *p == '\t')) break; 
			p++;
		}
		cmd = (char *)p;
		while (*p) {
			if (*p == ' ' || *p == '\t') break; 
			p++;
		}
		cmd_end = p;
		if (*p) p++;
		while (*p) {
			if (!(*p == ' ' || *p == '\t')) break; 
			p++;
		}
		if (*p == '>') {
			create_mode = 1;
			if (*p) p++;
			while (*p) {
				if (!(*p == ' ' || *p == '\t')) break; 
				p++;
			}
		}
		fname = p;
		while (*p) {
			if (*p == ' ' || *p == '\t') break; 
			p++;
		}
		*p = 0;
		*cmd_end = 0;
		if (!strcmp_P(cmd, PSTR("ls"))) {
			r = print_filelist(fname);
			if (r < 0) print_error(r);
		} else if (!strcmp_P(cmd, PSTR("cat"))) {
			if (create_mode) {
				r = create_file(fname);
			} else {
				r = print_file(fname);
			}
			if (r < 0) print_error(r);
		} else if (!strcmp_P(cmd, PSTR("rm"))) {
			r = remove_file(fname);
			if (r < 0) print_error(r);
		} else if (!strcmp_P(cmd, PSTR("passwd"))) {
			memset(ext_buf, 0xff , 16);
			eeprom_write_block(ext_buf, (void *)128, 16);
			login_ok = 0;
		} else if (!strcmp_P(cmd, PSTR("exit"))) {
#if defined(__linux__) || defined(WIN32NATIVE)
			exit(0);
#endif
			login_ok = 0;
#ifdef I2CROM_DEBUG
		} else if (!strcmp_P(cmd, PSTR("dump"))) {
			r = i2crom_dump(fname);
			if (r < 0) print_error(r);
#endif
		} else if (!strcmp_P(cmd, PSTR("format"))) {
			r = i2crom_format(fname);
			if (r < 0) print_error(r);
		} else if (cmd[0]) {
			puts_P(PSTR("command not found\r\n"));
		}
#endif
	}
	return 0;
}
